import logging


from PyQt5.QtCore import QModelIndex
from PyQt5.QtWidgets import (
    QAbstractItemView,
    QDialog,
    QFormLayout,
    QHBoxLayout,
    QLineEdit,
    QMessageBox,
    QPushButton,
    QWidget,
)

from ptychodus.api.observer import Observable, Observer
from ptychodus.api.parametric import PathParameter, StringParameter

from ...model.analysis import DiffractionSimulator
from ...model.diffraction import (
    AssembledDiffractionDataset,
    BadPixelsProvider,
    DetectorSettings,
    DiffractionAPI,
    DiffractionDatasetObserver,
    DiffractionSettings,
    PatternSizer,
)
from ...model.metadata import MetadataPresenter
from ...model.product import ProductRepository
from ...view.diffraction import DetectorView, PatternsView
from ...view.widgets import ExceptionDialog, ProgressBarItemDelegate
from ..data import FileDialogFactory
from ..helpers import connect_triggered_signal
from ..image import ImageController
from ..parametric import (
    LengthWidgetParameterViewController,
    ParameterViewController,
    SpinBoxParameterViewController,
)
from ..product.list_model import ProductRepositoryListModel
from .dataset import DatasetTreeModel
from .dataset_layout import DatasetLayoutViewController
from .wizard import OpenDatasetWizardController

logger = logging.getLogger(__name__)


class BadPixelsViewController(ParameterViewController, Observer):
    def __init__(
        self,
        bad_pixels_file_path: PathParameter,
        bad_pixels_file_type: StringParameter,
        bad_pixels_provider: BadPixelsProvider,
        diffraction_api: DiffractionAPI,
        file_dialog_factory: FileDialogFactory,
    ) -> None:
        super().__init__()
        self._bad_pixels_file_path = bad_pixels_file_path
        self._bad_pixels_file_type = bad_pixels_file_type
        self._bad_pixels_provider = bad_pixels_provider
        self._diffraction_api = diffraction_api
        self._file_dialog_factory = file_dialog_factory

        self._line_edit = QLineEdit()
        self._line_edit.setReadOnly(True)
        self._browse_button = QPushButton('Browse...')
        self._browse_button.clicked.connect(self._open_bad_pixels)
        self._clear_button = QPushButton('Clear')
        self._clear_button.clicked.connect(self._diffraction_api.clear_bad_pixels)
        self._widget = QWidget()

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._line_edit)
        layout.addWidget(self._browse_button)
        layout.addWidget(self._clear_button)
        self._widget.setLayout(layout)

        self._sync_model_to_view()
        bad_pixels_provider.add_observer(self)

    def _open_bad_pixels(self) -> None:
        file_reader_chooser = self._diffraction_api.get_bad_pixels_file_reader_chooser()
        current_plugin = file_reader_chooser.get_current_plugin()
        file_path, name_filter = self._file_dialog_factory.get_open_file_path(
            self._widget,
            'Open Bad Pixels File',
            name_filters=[plugin.display_name for plugin in file_reader_chooser],
            selected_name_filter=current_plugin.simple_name,
        )

        if file_path:
            try:
                self._diffraction_api.open_bad_pixels(file_path, file_type=name_filter)
            except Exception as exc:
                logger.exception(exc)
                ExceptionDialog.show_exception('Bad Pixels File Reader', exc)

    def get_widget(self) -> QWidget:
        return self._widget

    def _sync_model_to_view(self) -> None:
        num_bad_pixels = self._bad_pixels_provider.get_num_bad_pixels()
        self._line_edit.setText(str(num_bad_pixels))

    def _update(self, observable: Observable) -> None:
        if observable is self._bad_pixels_provider:
            self._sync_model_to_view()


class DetectorController:
    def __init__(
        self,
        settings: DetectorSettings,
        bad_pixels_provider: BadPixelsProvider,
        diffraction_api: DiffractionAPI,
        view: DetectorView,
        file_dialog_factory: FileDialogFactory,
    ) -> None:
        self._width_px_view_controller = SpinBoxParameterViewController(settings.width_px)
        self._height_px_view_controller = SpinBoxParameterViewController(settings.height_px)
        self._pixel_width_view_controller = LengthWidgetParameterViewController(
            settings.pixel_width_m
        )
        self._pixel_height_view_controller = LengthWidgetParameterViewController(
            settings.pixel_height_m
        )
        self._bit_depth_view_controller = SpinBoxParameterViewController(settings.bit_depth)
        self._bad_pixels_view_controller = BadPixelsViewController(
            settings.bad_pixels_file_path,
            settings.bad_pixels_file_type,
            bad_pixels_provider,
            diffraction_api,
            file_dialog_factory,
        )

        layout = QFormLayout()
        layout.addRow('Detector Width [px]:', self._width_px_view_controller.get_widget())
        layout.addRow('Detector Height [px]:', self._height_px_view_controller.get_widget())
        layout.addRow('Pixel Width:', self._pixel_width_view_controller.get_widget())
        layout.addRow('Pixel Height:', self._pixel_height_view_controller.get_widget())
        layout.addRow('Bit Depth:', self._bit_depth_view_controller.get_widget())
        layout.addRow('Bad Pixels:', self._bad_pixels_view_controller.get_widget())
        view.setLayout(layout)


class DiffractionController(DiffractionDatasetObserver):
    def __init__(
        self,
        detector_settings: DetectorSettings,
        diffraction_settings: DiffractionSettings,
        pattern_sizer: PatternSizer,
        bad_pixels_provider: BadPixelsProvider,
        diffraction_api: DiffractionAPI,
        dataset: AssembledDiffractionDataset,
        metadata_presenter: MetadataPresenter,
        product_repository: ProductRepository,
        diffraction_simulator: DiffractionSimulator,
        view: PatternsView,
        image_controller: ImageController,
        file_dialog_factory: FileDialogFactory,
    ) -> None:
        super().__init__()
        self._pattern_sizer = pattern_sizer
        self._diffraction_api = diffraction_api
        self._dataset = dataset
        self._product_list_model = ProductRepositoryListModel(product_repository)
        self._diffraction_simulator = diffraction_simulator
        self._view = view
        self._image_controller = image_controller
        self._file_dialog_factory = file_dialog_factory
        self._detector_controller = DetectorController(
            detector_settings,
            bad_pixels_provider,
            diffraction_api,
            view.detector_view,
            file_dialog_factory,
        )
        self._wizard_controller = OpenDatasetWizardController(
            diffraction_settings,
            pattern_sizer,
            diffraction_api,
            metadata_presenter,
            file_dialog_factory,
        )
        self._tree_model = DatasetTreeModel()

        view.tree_view.setModel(self._tree_model)
        view.tree_view.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        counts_item_delegate = ProgressBarItemDelegate(view.tree_view)
        view.tree_view.setItemDelegateForColumn(1, counts_item_delegate)
        selection_model = view.tree_view.selectionModel()

        if selection_model is None:
            raise ValueError('selection_model is None!')
        else:
            selection_model.currentChanged.connect(self._update_view)

        self._update_view(QModelIndex(), QModelIndex())

        open_dataset_action = view.button_box.load_menu.addAction('Open File...')
        connect_triggered_signal(open_dataset_action, self._wizard_controller.open_dataset)

        view.button_box.save_button.clicked.connect(self._save_dataset)
        view.button_box.close_button.clicked.connect(self._close_dataset)

        dataset_layout_action = view.button_box.analyze_menu.addAction('Dataset Layout...')
        connect_triggered_signal(dataset_layout_action, self._show_dataset_layout)
        simulate_action = view.button_box.analyze_menu.addAction('Simulate Diffraction...')
        connect_triggered_signal(simulate_action, self._choose_product_for_simulation)

        view.simulate_dialog.product_combo_box.setModel(self._product_list_model)
        view.simulate_dialog.finished.connect(self._simulate_diffraction)

        dataset.add_observer(self)
        self._sync_model_to_view()

    def _update_view(self, current: QModelIndex, previous: QModelIndex) -> None:
        if current.isValid():
            node = current.internalPointer()
            data = node.get_data()
            pixel_geometry = self._pattern_sizer.get_processed_pixel_geometry()
            self._image_controller.set_array(data, pixel_geometry)
        else:
            self._image_controller.clear_array()

    def _save_dataset(self) -> None:
        file_writer_chooser = self._diffraction_api.get_file_writer_chooser()
        file_path, name_filter = self._file_dialog_factory.get_save_file_path(
            self._view,
            'Save Diffraction File',
            name_filters=[plugin.display_name for plugin in file_writer_chooser],
            selected_name_filter=file_writer_chooser.get_current_plugin().display_name,
        )

        if file_path:
            try:
                self._diffraction_api.save_patterns(file_path, name_filter)
            except Exception as exc:
                logger.exception(exc)
                ExceptionDialog.show_exception('File Writer', exc)

    def _show_dataset_layout(self) -> None:
        DatasetLayoutViewController.show_dialog(self._dataset, self._view)

    def _choose_product_for_simulation(self) -> None:
        self._product_list_model.beginResetModel()
        self._product_list_model.endResetModel()
        self._view.simulate_dialog.open()

    def _simulate_diffraction(self, result: int) -> None:
        if result == QDialog.DialogCode.Accepted:
            item_index = self._view.simulate_dialog.product_combo_box.currentIndex()
            self._diffraction_simulator.simulate(item_index)

    def _close_dataset(self) -> None:
        button = QMessageBox.question(
            self._view,
            'Confirm Close',
            'This will free the diffraction data from memory. Do you want to continue?',
        )

        if button == QMessageBox.StandardButton.Yes:
            self._diffraction_api.close_patterns()
            self._update_view(QModelIndex(), QModelIndex())

    def _sync_model_to_view(self) -> None:
        self._tree_model.clear()

        for index, array in enumerate(self._dataset):
            self._tree_model.insert_array(index, array)  # type: ignore

        info_text = self._dataset.get_info_text()
        self._view.info_label.setText(info_text)

    def handle_array_inserted(self, index: int) -> None:
        self._tree_model.insert_array(index, self._dataset[index])

    def handle_array_changed(self, index: int) -> None:
        self._tree_model.refresh_array(index)

    def handle_dataset_reloaded(self) -> None:
        self._sync_model_to_view()



from __future__ import annotations

from PyQt5.QtWidgets import QVBoxLayout, QWidget

from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg
from matplotlib.backends.backend_qt import NavigationToolbar2QT as NavigationToolbar
from matplotlib.figure import Figure


class ProbePositionsPlotView(QWidget):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.figure = Figure()
        self.figure_canvas = FigureCanvasQTAgg(self.figure)
        self.navigation_toolbar = NavigationToolbar(self.figure_canvas, self)
        self.axes = self.figure.add_subplot(111)

        layout = QVBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.navigation_toolbar)
        layout.addWidget(self.figure_canvas)
        self.setLayout(layout)



from pathlib import Path

import numpy

from ptychodus.api.object import Object, ObjectFileReader, ObjectFileWriter
from ptychodus.api.plugins import PluginRegistry


class CSVObjectFileReader(ObjectFileReader):
    def read(self, file_path: Path) -> Object:
        array = numpy.genfromtxt(file_path, delimiter=',', dtype=complex)
        return Object(array=array, pixel_geometry=None, center=None)


class CSVObjectFileWriter(ObjectFileWriter):
    def write(self, file_path: Path, object_: Object) -> None:
        array = object_.get_array()
        numpy.savetxt(file_path, array, delimiter=',')


def register_plugins(registry: PluginRegistry) -> None:
    registry.object_file_readers.register_plugin(
        CSVObjectFileReader(),
        simple_name='CSV',
        display_name='Comma-Separated Values Files (*.csv)',
    )
    registry.object_file_writers.register_plugin(
        CSVObjectFileWriter(),
        simple_name='CSV',
        display_name='Comma-Separated Values Files (*.csv)',
    )



from PyQt5.QtWidgets import QWidget

from ..model.ptychonn import PtychoNNReconstructorLibrary
from .parametric import ParameterViewBuilder
from .processing import ReconstructorViewControllerFactory


class PtychoNNViewControllerFactory(ReconstructorViewControllerFactory):
    def __init__(self, model: PtychoNNReconstructorLibrary) -> None:
        super().__init__()
        self._model = model

    @property
    def name(self) -> str:
        return 'PtychoNN'

    def create_view_controller(self, reconstructor_name: str) -> QWidget:
        view_builder = ParameterViewBuilder()

        model_settings = self._model.model_settings
        model_group = 'Model Parameters'
        view_builder.add_spin_box(
            model_settings.num_convolution_kernels, 'Convolution Kernels:', group=model_group
        )
        view_builder.add_spin_box(model_settings.batch_size, 'Batch Size:', group=model_group)
        view_builder.add_check_box(
            model_settings.use_batch_normalization, 'Use Batch Normalization', group=model_group
        )

        training_settings = self._model.training_settings
        training_group = 'Training Parameters'

        view_builder.add_decimal_slider(
            training_settings.validation_set_fractional_size,
            'Validation Set Fractional Size:',
            group=training_group,
        )
        view_builder.add_decimal_line_edit(
            training_settings.max_learning_rate, 'Max Learning Rate:', group=training_group
        )
        view_builder.add_decimal_line_edit(
            training_settings.min_learning_rate, 'Min Learning Rate:', group=training_group
        )
        view_builder.add_spin_box(
            training_settings.training_epochs, 'Training Epochs:', group=training_group
        )
        view_builder.add_spin_box(
            training_settings.status_interval_in_epochs, 'Status Interval:', group=training_group
        )

        return view_builder.build_widget()



from __future__ import annotations
from itertools import pairwise
import logging

import numpy

from ptychodus.api.observer import Observable
from ptychodus.api.parametric import ParameterGroup
from ptychodus.api.probe_positions import ProbePositionSequence, ScanBoundingBox, ProbePosition

from .bounding_box import ScanBoundingBoxBuilder
from .builder import FromMemoryProbePositionsBuilder, ProbePositionsBuilder
from .settings import ProbePositionsSettings
from .transform import ProbePositionTransform

logger = logging.getLogger(__name__)


class ProbePositionsRepositoryItem(ParameterGroup):
    def __init__(
        self,
        settings: ProbePositionsSettings,
        builder: ProbePositionsBuilder,
        transform: ProbePositionTransform,
    ) -> None:
        super().__init__()
        self._settings = settings
        self._builder = builder
        self._transform = transform
        self._untransformed_scan = ProbePositionSequence()
        self._transformed_scan = ProbePositionSequence()
        self._bbox_builder = ScanBoundingBoxBuilder()
        self._length_m = 0.0

        self._add_group('builder', builder, observe=True)
        self._add_group('transform', transform, observe=True)

        self.expand_bbox = settings.expand_bbox.copy()
        self._add_parameter('expand_bbox', self.expand_bbox)

        self.expand_bbox_xmin_m = settings.expand_bbox_xmin_m.copy()
        self._add_parameter('expand_bbox_xmin_m', self.expand_bbox_xmin_m)

        self.expand_bbox_xmax_m = settings.expand_bbox_xmax_m.copy()
        self._add_parameter('expand_bbox_xmax_m', self.expand_bbox_xmax_m)

        self.expand_bbox_ymin_m = settings.expand_bbox_ymin_m.copy()
        self._add_parameter('expand_bbox_ymin_m', self.expand_bbox_ymin_m)

        self.expand_bbox_ymax_m = settings.expand_bbox_ymax_m.copy()
        self._add_parameter('expand_bbox_ymax_m', self.expand_bbox_ymax_m)

        self._rebuild()

    def assign_item(self, item: ProbePositionsRepositoryItem) -> None:
        group = 'transform'

        self._remove_group(group)
        self._transform.remove_observer(self)

        transform = item.get_transform()

        self._transform = transform.copy()
        self._transform.add_observer(self)
        self._add_group(group, self._transform, observe=True)

        self.set_builder(item.get_builder().copy())
        self._rebuild()

    def assign(self, scan: ProbePositionSequence) -> None:
        builder = FromMemoryProbePositionsBuilder(self._settings, scan)
        self.set_builder(builder)

    def sync_to_settings(self) -> None:
        for parameter in self.parameters().values():
            parameter.sync_value_to_parent()

        self._builder.sync_to_settings()
        self._transform.sync_to_settings()

    def get_probe_positions(self) -> ProbePositionSequence:
        return self._transformed_scan

    def get_builder(self) -> ProbePositionsBuilder:
        return self._builder

    def set_builder(self, builder: ProbePositionsBuilder) -> None:
        group = 'builder'
        self._remove_group(group)
        self._builder.remove_observer(self)
        self._builder = builder
        self._builder.add_observer(self)
        self._add_group(group, self._builder, observe=True)
        self._rebuild()

    def get_bounding_box(self) -> ScanBoundingBox | None:
        bbox = self._bbox_builder.get_bounding_box()

        if self.expand_bbox.get_value():
            expanded_bbox = ScanBoundingBox(
                minimum_x_m=self.expand_bbox_xmin_m.get_value(),
                maximum_x_m=self.expand_bbox_xmax_m.get_value(),
                minimum_y_m=self.expand_bbox_ymin_m.get_value(),
                maximum_y_m=self.expand_bbox_ymax_m.get_value(),
            )
            bbox = expanded_bbox if bbox is None else bbox.hull(expanded_bbox)

        return bbox

    def get_length_m(self) -> float:
        return self._length_m

    def _transform_scan(self) -> None:
        transformed_points: list[ProbePosition] = list()
        bbox_builder = ScanBoundingBoxBuilder()
        length_m = 0.0

        for untransformed_point in self._untransformed_scan:
            transformed_point = self._transform(untransformed_point)
            transformed_points.append(transformed_point)
            bbox_builder.hull(transformed_point)

        for point_l, point_r in pairwise(transformed_points):
            dx = point_r.coordinate_x_m - point_l.coordinate_x_m
            dy = point_r.coordinate_y_m - point_l.coordinate_y_m
            length_m += numpy.hypot(dx, dy)

        self._transformed_scan = ProbePositionSequence(transformed_points)
        self._bbox_builder = bbox_builder
        self._length_m = length_m
        self.notify_observers()

    def _rebuild(self) -> None:
        try:
            scan = self._builder.build()
        except Exception:
            logger.exception('Failed to rebuild scan!')
            return

        self._untransformed_scan = scan
        self._transform_scan()

    def get_transform(self) -> ProbePositionTransform:
        return self._transform

    def _update(self, observable: Observable) -> None:
        if observable is self._builder:
            self._rebuild()
        elif observable is self._transform:
            self._transform_scan()
        else:
            super()._update(observable)



from pathlib import Path
from typing import Final

import h5py
import numpy

from ptychodus.api.common import RealArrayType
from ptychodus.api.fluorescence import (
    ElementMap,
    FluorescenceDataset,
    FluorescenceFileReader,
    FluorescenceFileWriter,
)
from ptychodus.api.plugins import PluginRegistry


class XRFMapsFileIO(FluorescenceFileReader, FluorescenceFileWriter):
    SIMPLE_NAME: Final[str] = 'XRF-Maps'
    DISPLAY_NAME: Final[str] = 'XRF-Maps Fluorescence Dataset (*.h5 *.h5*)'

    @staticmethod
    def _split_path(data_path: str) -> tuple[str, str]:
        parts = data_path.split('/')
        return '/'.join(parts[:-1]), parts[-1]

    def read(self, file_path: Path) -> FluorescenceDataset:
        element_maps: list[ElementMap] = list()
        counts_per_second_path = str()
        channel_names_path = str()

        with h5py.File(file_path, 'r') as h5_file:
            # try to see if v10 layout, Non Negative Lease squares fitting tech was used
            h5_counts_per_second = h5_file['/MAPS/XRF_Analyzed/NNLS/Counts_Per_Sec']
            h5_channel_names = h5_file['/MAPS/XRF_Analyzed/NNLS/Channel_Names']

            if h5_counts_per_second is None:
                # try to see if v10 layout, iterative matrix fitting tech was used
                h5_counts_per_second = h5_file['/MAPS/XRF_Analyzed/Fitted/Counts_Per_Sec']
                h5_channel_names = h5_file['/MAPS/XRF_Analyzed/Fitted/Channel_Names']

                if h5_counts_per_second is None:
                    # try to see if was saved in v9 layout
                    h5_counts_per_second = h5_file['/MAPS/XRF_fits']
                    h5_channel_names = h5_file['/MAPS/channel_names']

            if h5_counts_per_second is not None:
                # Counts_Per_Sec is an N x H x W
                # where N is number of elements, use channel_names to find what element index
                counts_per_second = h5_counts_per_second[...]
                channel_names = h5_channel_names[...]

                for bname, cps in zip(channel_names, counts_per_second):
                    string_info = h5py.check_string_dtype(bname.dtype)
                    name = bname.decode(string_info.encoding)
                    emap = ElementMap(name, cps)
                    element_maps.append(emap)

            counts_per_second_path = h5_counts_per_second.name
            channel_names_path = h5_channel_names.name

        return FluorescenceDataset(
            element_maps=element_maps,
            counts_per_second_path=counts_per_second_path,
            channel_names_path=channel_names_path,
        )

    def write(self, file_path: Path, dataset: FluorescenceDataset) -> None:
        channel_names: list[str] = list()
        counts_per_sec: list[RealArrayType] = list()

        for emap in dataset.element_maps:
            channel_names.append(emap.name)
            counts_per_sec.append(emap.counts_per_second)

        cps_group_path, cps_dataset_name = self._split_path(dataset.counts_per_second_path)
        ch_group_path, ch_dataset_name = self._split_path(dataset.channel_names_path)

        with h5py.File(file_path, 'w') as h5_file:
            cps_group = h5_file.require_group(cps_group_path)
            cps_group.create_dataset(cps_dataset_name, data=numpy.stack(counts_per_sec))
            ch_group = h5_file.require_group(ch_group_path)
            ch_group.create_dataset(ch_dataset_name, data=channel_names, dtype='S256')


class NPZFluorescenceFileWriter(FluorescenceFileWriter):
    def write(self, file_path: Path, dataset: FluorescenceDataset) -> None:
        contents = {emap.name: emap.counts_per_second for emap in dataset.element_maps}
        numpy.savez_compressed(file_path, allow_pickle=False, **contents)


def register_plugins(registry: PluginRegistry) -> None:
    xrf_maps_file_io = XRFMapsFileIO()

    registry.fluorescence_file_readers.register_plugin(
        xrf_maps_file_io,
        simple_name=XRFMapsFileIO.SIMPLE_NAME,
        display_name=XRFMapsFileIO.DISPLAY_NAME,
    )
    registry.fluorescence_file_writers.register_plugin(
        xrf_maps_file_io,
        simple_name=XRFMapsFileIO.SIMPLE_NAME,
        display_name=XRFMapsFileIO.DISPLAY_NAME,
    )
    registry.fluorescence_file_writers.register_plugin(
        NPZFluorescenceFileWriter(),
        simple_name='NPZ',
        display_name='NumPy Zipped Archive (*.npz)',
    )



from collections.abc import Iterator
from typing import Final

from matplotlib.colors import Colormap
import matplotlib

from ptychodus.api.observer import Observable, Observer
from ptychodus.api.parametric import Parameter
from ptychodus.api.plugins import PluginChooser


class ColormapParameter(Parameter[str], Observer):
    # See https://matplotlib.org/stable/gallery/color/colormap_reference.html
    CYCLIC_COLORMAPS: Final[tuple[str, ...]] = ('hsv', 'twilight', 'twilight_shifted')

    def __init__(self, *, is_cyclic: bool) -> None:
        super().__init__()
        self._is_cyclic = is_cyclic
        self._chooser = PluginChooser[Colormap]()

        for name, cmap in matplotlib.colormaps.items():
            is_cyclic_colormap = name in ColormapParameter.CYCLIC_COLORMAPS

            if is_cyclic == is_cyclic_colormap:
                self._chooser.register_plugin(cmap, display_name=name)

        self.set_value('hsv' if is_cyclic else 'gray')
        self._chooser.add_observer(self)

    def choices(self) -> Iterator[str]:
        for plugin in self._chooser:
            yield plugin.display_name

    def get_value(self) -> str:
        return self._chooser.get_current_plugin().display_name

    def set_value(self, value: str, *, notify: bool = True) -> None:
        self._chooser.set_current_plugin(value)

    def get_value_as_string(self) -> str:
        return self.get_value()

    def set_value_from_string(self, value: str) -> None:
        self.set_value(value)

    def copy(self) -> Parameter[str]:
        parameter = ColormapParameter(is_cyclic=self._is_cyclic)
        parameter.set_value(self.get_value())
        return parameter

    def get_plugin(self) -> Colormap:
        return self._chooser.get_current_plugin().strategy

    def _update(self, observable: Observable) -> None:
        if observable is self._chooser:
            self.notify_observers()



#!/usr/bin/env python
"""
Prepare experiment data for use in a beamline data pipeline
"""

from pathlib import Path
import argparse
import logging
import sys

from ptychodus.api.diffraction import CropCenter
from ptychodus.api.geometry import ImageExtent
from ptychodus.api.io import StandardFileLayout
from ptychodus.cli import DirectoryType
from ptychodus.model import ModelCore
import ptychodus

logger = logging.getLogger(__name__)


def main() -> int:
    crop_center: CropCenter | None = None
    crop_extent: ImageExtent | None = None

    prog = Path(__file__).stem.lower()
    parser = argparse.ArgumentParser(
        prog=prog,
        description=f'{prog} prepares experiment data for use in beamline data pipelines',
    )
    parser.add_argument(
        '--crop-center-x-px',
        metavar='CENTER_X',
        help='Diffraction pattern crop center x in pixels',
        type=int,
    )
    parser.add_argument(
        '--crop-center-y-px',
        metavar='CENTER_Y',
        help='Diffraction pattern crop center y in pixels',
        type=int,
    )
    parser.add_argument(
        '--crop-width-px',
        metavar='WIDTH',
        help='Diffraction pattern crop width in pixels',
        type=int,
    )
    parser.add_argument(
        '--crop-height-px',
        metavar='HEIGHT',
        help='Diffraction pattern crop height in pixels',
        type=int,
    )
    parser.add_argument(
        '--defocus-distance-m',
        metavar='DISTANCE',
        help='Defocus distance in meters',
        type=float,
    )
    parser.add_argument(
        '--detector-distance-m',
        metavar='DISTANCE',
        help='Detector distance in meters',
        type=float,
    )
    parser.add_argument(
        '--diffraction-input',
        metavar='DIFFRACTION_INPUT_FILE',
        help='Path to diffraction file.',
        type=argparse.FileType('r'),
        required=True,
    )
    parser.add_argument(
        '--exposure-time-s',
        metavar='TIME',
        help='Exposure time in seconds',
        type=float,
    )
    parser.add_argument(
        '--log-level',
        default=logging.INFO,
        help='Python logging level.',
        type=int,
    )
    parser.add_argument(
        '-o',
        '--output-directory',
        metavar='OUTPUT_DIR',
        type=DirectoryType(must_exist=False),
        required=True,
    )
    parser.add_argument(
        '--probe-energy-eV',
        metavar='ENERGY',
        help='Probe energy in electron volts',
        type=float,
    )
    parser.add_argument(
        '--probe-photon-count',
        metavar='NUMBER',
        help='Probe number of photons',
        type=float,
    )
    parser.add_argument(
        '--probe-position-input',
        metavar='PROBE_POSITION_INPUT_FILE',
        help='Path to probe position input file',
        type=argparse.FileType('r'),
        required=True,
    )
    parser.add_argument(
        '--product-comment',
        default='',
        help='Data product comment',
    )
    parser.add_argument(
        '--product-name',
        help='Data product name',
        required=True,
    )
    parser.add_argument(
        '-s',
        '--settings',
        metavar='SETTINGS_FILE',
        help='Use default settings from file',
        type=argparse.FileType('r'),
        required=True,
    )
    parser.add_argument(
        '--tomography-angle-deg',
        metavar='ANGLE',
        help='Tomography angle in degrees',
        type=float,
    )
    parser.add_argument(
        '-v',
        '--version',
        action='version',
        version=ptychodus.VERSION_STRING,
    )

    args = parser.parse_args()

    if args.crop_center_x_px is not None and args.crop_center_y_px is not None:
        crop_center = CropCenter(
            position_x_px=args.crop_center_x_px,
            position_y_px=args.crop_center_y_px,
        )
    elif bool(args.crop_center_x_px) ^ bool(args.crop_center_y_px):
        parser.error('--crop-center-x-px and --crop-center-y-px must be given together.')

    if args.crop_width_px is not None and args.crop_height_px is not None:
        crop_extent = ImageExtent(
            width_px=args.crop_width_px,
            height_px=args.crop_height_px,
        )
    elif bool(args.crop_width_px) ^ bool(args.crop_height_px):
        parser.error('--crop-width-px and --crop-height-px must be given together.')

    if args.defocus_distance_m is not None:
        logger.warning('Defocus distance is not implemented yet!')  # TODO

    with ModelCore(Path(args.settings.name), log_level=args.log_level) as model:
        workflow_diffraction_api = model.workflow_api.load_diffraction_data(
            Path(args.diffraction_input.name),
            crop_center=crop_center,
            crop_extent=crop_extent,
            block=True,
        )
        workflow_product_api = model.workflow_api.create_product(
            name=args.product_name,
            comments=args.product_comment,
            detector_distance_m=args.detector_distance_m,
            probe_energy_eV=args.probe_energy_eV,
            probe_photon_count=args.probe_photon_count,
            exposure_time_s=args.exposure_time_s,
            tomography_angle_deg=args.tomography_angle_deg,
        )
        workflow_product_api.load_probe_positions(Path(args.probe_position_input.name))
        workflow_product_api.generate_probe()
        workflow_product_api.generate_object()

        staging_dir = args.output_directory
        staging_dir.mkdir(parents=True, exist_ok=True)
        workflow_diffraction_api.save_assembled_data(staging_dir / StandardFileLayout.DIFFRACTION)
        workflow_product_api.save_product(
            staging_dir / StandardFileLayout.PRODUCT_IN, file_type='HDF5'
        )
        model.workflow_api.save_settings(staging_dir / StandardFileLayout.SETTINGS)

    return 0


if __name__ == '__main__':
    sys.exit(main())



from pathlib import Path
import logging

import h5py
import numpy

from ptychodus.api.geometry import ImageExtent
from ptychodus.api.diffraction import (
    DiffractionDataset,
    DiffractionFileReader,
    DiffractionMetadata,
    SimpleDiffractionDataset,
)
from ptychodus.api.plugins import PluginRegistry

from .h5_diffraction_file import H5DiffractionPatternArray, H5DiffractionFileTreeBuilder

logger = logging.getLogger(__name__)


class PolarDiffractionFileReader(DiffractionFileReader):
    def __init__(self) -> None:
        self._data_path = '/entry/externals/eiger'
        self._tree_builder = H5DiffractionFileTreeBuilder()

    def read(self, file_path: Path) -> DiffractionDataset:
        with h5py.File(file_path, 'r') as h5_file:
            contents_tree = self._tree_builder.build(h5_file)
            data_link = h5_file.get(self._data_path, getlink=True)

        if not isinstance(data_link, h5py.ExternalLink):
            raise ValueError(
                f'Expected "{self._data_path}" to be an external link; got {type(data_link)}.'
            )

        data_file_path = file_path.parent / data_link.filename
        logger.debug(f'Opening "{data_file_path}"...')

        with h5py.File(data_file_path, 'r') as h5_file:
            data = h5_file[data_link.path]

            if isinstance(data, h5py.Group):
                logger.warning('Link points to group; falling back to "/entry/data/data"')
                data = h5_file['/entry/data/data']

            if isinstance(data, h5py.Dataset):
                num_patterns, detector_height, detector_width = data.shape

                metadata = DiffractionMetadata(
                    num_patterns_per_array=[num_patterns],
                    pattern_dtype=data.dtype,
                    detector_extent=ImageExtent(detector_width, detector_height),
                    file_path=file_path,
                )

                array = H5DiffractionPatternArray(
                    label=file_path.stem,
                    indexes=numpy.arange(num_patterns),
                    file_path=data_file_path,
                    data_path=data.name,
                )
            else:
                raise ValueError(f'Expected "{data.name}" to be a dataset; got {type(data)}.')

        return SimpleDiffractionDataset(metadata, contents_tree, [array])


def register_plugins(registry: PluginRegistry) -> None:
    registry.diffraction_file_readers.register_plugin(
        PolarDiffractionFileReader(),
        simple_name='APS_Polar',
        display_name='APS 4-ID Polar Files (*.hdf)',
    )



from __future__ import annotations

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QLabel,
    QLineEdit,
    QVBoxLayout,
    QWidget,
)

from ...model.globus import GlobusAuthorizer


class GlobusAuthorizationDialog(QDialog):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.label = QLabel()
        self.line_edit = QLineEdit()
        self.button_box = QDialogButtonBox()

        self.label.setTextFormat(Qt.TextFormat.RichText)
        self.label.setTextInteractionFlags(Qt.TextInteractionFlag.TextBrowserInteraction)
        self.label.setOpenExternalLinks(True)

        self.button_box.addButton(QDialogButtonBox.StandardButton.Ok)
        self.button_box.accepted.connect(self.accept)
        self.button_box.addButton(QDialogButtonBox.StandardButton.Cancel)
        self.button_box.rejected.connect(self.reject)

        self.setWindowTitle('Authorize Workflow')

        layout = QVBoxLayout()
        layout.addWidget(self.label)
        layout.addWidget(self.line_edit)
        layout.addWidget(self.button_box)
        self.setLayout(layout)


class GlobusAuthorizationController:
    def __init__(self, authorizer: GlobusAuthorizer, dialog_parent: QWidget) -> None:
        super().__init__()
        self._authorizer = authorizer
        self._dialog = GlobusAuthorizationDialog(dialog_parent)

        self._dialog.finished.connect(self._finish_authorization)
        self._dialog.line_edit.textChanged.connect(self._set_dialog_buttons_enabled)
        self._set_dialog_buttons_enabled()

    def _set_dialog_buttons_enabled(self) -> None:
        ok_button = self._dialog.button_box.button(QDialogButtonBox.StandardButton.Ok)
        ok_button.setEnabled(len(self._dialog.line_edit.text()) > 0)

    def authorize_as_needed(self) -> None:
        if self._authorizer.needs_authorize_code and not self._dialog.isVisible():
            self._start_authorization()

    def _start_authorization(self) -> None:
        authorize_url = self._authorizer.get_authorize_url()
        text = f'Input the Globus authorization code from <a href="{authorize_url}">this link</a>:'

        self._dialog.label.setText(text)
        self._dialog.line_edit.clear()
        self._dialog.setModal(True)
        self._dialog.open()

    def _finish_authorization(self, result: int) -> None:
        if result == QDialog.DialogCode.Accepted:
            auth_code = self._dialog.line_edit.text()
            self._authorizer.set_code_from_authorize_url(auth_code)
        else:
            self._authorizer.cancel_authorization()



from pathlib import Path
from typing import Final
import logging

import h5py
import numpy

from ptychodus.api.geometry import ImageExtent, PixelGeometry
from ptychodus.api.diffraction import (
    DiffractionDataset,
    DiffractionFileReader,
    DiffractionMetadata,
    SimpleDiffractionDataset,
)
from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.product import (
    ELECTRON_VOLT_J,
    LIGHT_SPEED_M_PER_S,
    PLANCK_CONSTANT_J_PER_HZ,
)

from .h5_diffraction_file import H5DiffractionPatternArray, H5DiffractionFileTreeBuilder

logger = logging.getLogger(__name__)


class NSLS2Style1DiffractionFileReader(DiffractionFileReader):
    def __init__(self) -> None:
        self._tree_builder = H5DiffractionFileTreeBuilder()

    def read(self, file_path: Path) -> DiffractionDataset:
        with h5py.File(file_path, 'r') as h5_file:
            try:
                data = h5_file['/entry/data/data']
            except KeyError:
                data = h5_file['/diff_data/merlin1/det_images']

            if isinstance(data, h5py.Dataset):
                num_patterns, detector_height, detector_width = data.shape

                metadata = DiffractionMetadata(
                    num_patterns_per_array=[num_patterns],
                    pattern_dtype=data.dtype,
                    detector_extent=ImageExtent(detector_width, detector_height),
                    probe_energy_eV=float(h5_file['/scan/energy'][()]),
                    exposure_time_s=float(h5_file['/scan/exposure_time'][()]),
                    file_path=file_path,
                )
                contents_tree = self._tree_builder.build(h5_file)
                array = H5DiffractionPatternArray(
                    label=file_path.stem,
                    indexes=numpy.arange(num_patterns),
                    file_path=file_path,
                    data_path=data.name,
                )

                return SimpleDiffractionDataset(metadata, contents_tree, [array])
            else:
                raise ValueError(f'Expected {data.name} to be a dataset; got {type(data)}.')


class NSLS2Style2DiffractionFileReader(DiffractionFileReader):
    ONE_MICRON_M: Final[float] = 1.0e-6
    ONE_NANOMETER_M: Final[float] = 1.0e-9

    def __init__(self) -> None:
        self._tree_builder = H5DiffractionFileTreeBuilder()

    def read(self, file_path: Path) -> DiffractionDataset:
        with h5py.File(file_path, 'r') as h5_file:
            data = h5_file['/diffamp']

            if isinstance(data, h5py.Dataset):
                num_patterns, detector_height, detector_width = data.shape

                pixel_size_m = float(h5_file['/ccd_pixel_um'][()]) * self.ONE_MICRON_M
                wavelength_m = float(h5_file['/lambda_nm'][()]) * self.ONE_NANOMETER_M
                hc_Jm = PLANCK_CONSTANT_J_PER_HZ * LIGHT_SPEED_M_PER_S  # noqa: N806
                hc_eVm = hc_Jm / ELECTRON_VOLT_J  # noqa: N806
                probe_energy_eV = hc_eVm / wavelength_m  # noqa: N806

                metadata = DiffractionMetadata(
                    num_patterns_per_array=[num_patterns],
                    pattern_dtype=data.dtype,
                    detector_distance_m=float(h5_file['z_m'][()]),
                    detector_extent=ImageExtent(detector_width, detector_height),
                    detector_pixel_geometry=PixelGeometry(pixel_size_m, pixel_size_m),
                    probe_energy_eV=probe_energy_eV,
                    tomography_angle_deg=float(h5_file['angle'][()]),
                    file_path=file_path,
                )
                contents_tree = self._tree_builder.build(h5_file)
                array = H5DiffractionPatternArray(
                    label=file_path.stem,
                    indexes=numpy.arange(num_patterns),
                    file_path=file_path,
                    data_path=data.name,
                )

                return SimpleDiffractionDataset(metadata, contents_tree, [array])
            else:
                raise ValueError(f'Expected {data.name} to be a dataset; got {type(data)}.')


class NSLS2MATLABDiffractionFileReader(DiffractionFileReader):
    ONE_MICRON_M: Final[float] = 1e-6

    def __init__(self) -> None:
        self._tree_builder = H5DiffractionFileTreeBuilder()

    def read(self, file_path: Path) -> DiffractionDataset:
        with h5py.File(file_path, 'r') as h5_file:
            data = h5_file['det_data']

            if isinstance(data, h5py.Dataset):
                num_patterns, detector_height, detector_width = data.shape
                pixel_size_m = (
                    float(numpy.squeeze(h5_file['det_pixel_size'][()])) * self.ONE_MICRON_M
                )

                metadata = DiffractionMetadata(
                    num_patterns_per_array=[num_patterns],
                    pattern_dtype=data.dtype,
                    detector_extent=ImageExtent(detector_width, detector_height),
                    detector_pixel_geometry=PixelGeometry(pixel_size_m, pixel_size_m),
                    file_path=file_path,
                )
                contents_tree = self._tree_builder.build(h5_file)
                array = H5DiffractionPatternArray(
                    label=file_path.stem,
                    indexes=numpy.arange(num_patterns),
                    file_path=file_path,
                    data_path=data.name,
                )

                return SimpleDiffractionDataset(metadata, contents_tree, [array])
            else:
                raise ValueError(f'Expected {data.name} to be a dataset; got {type(data)}.')


def register_plugins(registry: PluginRegistry) -> None:
    registry.diffraction_file_readers.register_plugin(
        NSLS2Style1DiffractionFileReader(),
        simple_name='NSLS_II_1',
        display_name='NSLS-II Style 1 Files (*.h5 *.hdf5)',
    )
    registry.diffraction_file_readers.register_plugin(
        NSLS2Style2DiffractionFileReader(),
        simple_name='NSLS_II_2',
        display_name='NSLS-II Style 2 Files (*.h5 *.hdf5)',
    )
    registry.diffraction_file_readers.register_plugin(
        NSLS2MATLABDiffractionFileReader(),
        simple_name='NSLS_II_MATLAB',
        display_name='NSLS-II MATLAB Files (*.mat)',
    )



from __future__ import annotations
from collections import OrderedDict
from dataclasses import dataclass
from pathlib import Path
from time import monotonic as time
from typing import Final
import logging
import threading

from ptychodus.api.observer import Observable, Observer
from ptychodus.api.plugins import PluginChooser
from ptychodus.api.workflow import FileBasedWorkflow, WorkflowAPI

from ..task_manager import TaskManager
from .repository import (
    AutomationDatasetRepository,
    AutomationDatasetState,
    UpdateAutomationDatasetRepository,
)
from .settings import AutomationSettings

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class FileStatus:
    stat_time_s: float
    mtime_ns: int
    size_bytes: int

    @classmethod
    def from_file(cls, file_path: Path) -> FileStatus:
        stat = file_path.stat()
        return cls(time(), stat.st_mtime_ns, stat.st_size)


class ExecuteWorkflow:
    def __init__(
        self,
        repository: AutomationDatasetRepository,
        workflow: FileBasedWorkflow,
        workflow_api: WorkflowAPI,
        file_path: Path,
    ) -> None:
        self._repository = repository
        self._workflow = workflow
        self._workflow_api = workflow_api
        self._file_path = file_path

    def __call__(self) -> None:
        self._repository.upsert(self._file_path, AutomationDatasetState.RUNNING)
        self._workflow.execute(self._workflow_api, self._file_path)
        self._repository.upsert(self._file_path, AutomationDatasetState.COMPLETE)


class PendingFileBuffer(Observable, Observer):
    WAIT_TIME_S: Final[int] = 5

    def __init__(
        self,
        task_manager: TaskManager,
        settings: AutomationSettings,
        repository: AutomationDatasetRepository,
        workflow_api: WorkflowAPI,
        workflow_chooser: PluginChooser[FileBasedWorkflow],
    ) -> None:
        super().__init__()
        self._task_manager = task_manager
        self._settings = settings
        self._repository = repository
        self._workflow_api = workflow_api
        self._workflow_chooser = workflow_chooser

        self._event_times_lock = threading.Lock()
        self._event_times: OrderedDict[Path, FileStatus] = OrderedDict()
        self._stop_event = threading.Event()
        self._worker: threading.Thread | None = None

        self._watchdog_delay_lock = threading.Lock()
        self._watchdog_delay_s = settings.watchdog_delay_s.get_value()

        self._current_workflow_plugin_lock = threading.Lock()
        self._current_workflow_plugin = workflow_chooser.get_current_plugin()

        settings.watchdog_delay_s.add_observer(self)
        workflow_chooser.add_observer(self)

    def get_current_workflow(self) -> FileBasedWorkflow:
        with self._current_workflow_plugin_lock:
            return self._current_workflow_plugin.strategy

    def put(self, file_path: Path) -> None:
        file_status = FileStatus.from_file(file_path)

        with self._event_times_lock:
            self._event_times[file_path] = file_status
            self._event_times.move_to_end(file_path)

        task = UpdateAutomationDatasetRepository(
            self._repository, file_path, AutomationDatasetState.FOUND
        )
        self._task_manager.put_foreground_task(task)

    def _process_finalized_files(self) -> None:
        while not self._stop_event.is_set():
            try:
                with self._event_times_lock:
                    file_path, last_file_status = self._event_times.popitem(last=False)
            except KeyError:
                pass
            else:
                elapsed_time_s = time() - last_file_status.stat_time_s

                with self._watchdog_delay_lock:
                    remaining_time_s = self._watchdog_delay_s - elapsed_time_s

                if remaining_time_s > 0.0:
                    self._stop_event.wait(timeout=remaining_time_s)

                file_status = FileStatus.from_file(file_path)
                same_size = file_status.size_bytes == last_file_status.size_bytes
                same_mtime = file_status.mtime_ns == last_file_status.mtime_ns

                if same_size and same_mtime:
                    update_repository_task = UpdateAutomationDatasetRepository(
                        self._repository, file_path, AutomationDatasetState.WAITING
                    )
                    self._task_manager.put_foreground_task(update_repository_task)

                    with self._current_workflow_plugin_lock:
                        current_workflow = self._current_workflow_plugin.strategy

                    execute_workflow_task = ExecuteWorkflow(
                        self._repository, current_workflow, self._workflow_api, file_path
                    )
                    self._task_manager.put_foreground_task(execute_workflow_task)

                continue

            self._stop_event.wait(timeout=PendingFileBuffer.WAIT_TIME_S)

    def start(self) -> None:
        if self._worker is None:
            logger.debug('Starting automation buffer...')
            self._stop_event.clear()
            self._worker = threading.Thread(target=self._process_finalized_files)
            self._worker.start()
            logger.debug('Automation buffer started.')
        else:
            logger.debug('Automation buffer already started!')

    def stop(self) -> None:
        if self._stop_event.is_set():
            logger.debug('Automation buffer already stopped.')
        elif self._worker is None:
            logger.warning('Worker is None!')
        else:
            logger.debug('Stopping automation buffer...')
            self._stop_event.set()
            self._worker.join()
            self._worker = None
            logger.debug('Automation buffer stopped.')

    def _update(self, observable: Observable) -> None:
        if observable is self._settings.watchdog_delay_s:
            with self._watchdog_delay_lock:
                self._watchdog_delay_s = self._settings.watchdog_delay_s.get_value()
        if observable is self._workflow_chooser:
            with self._current_workflow_plugin_lock:
                self._current_workflow_plugin = self._workflow_chooser.get_current_plugin()

            self.notify_observers()



from PyQt5.QtWidgets import QFormLayout

from ptychodus.api.parametric import (
    BooleanParameter,
    IntegerParameter,
    RealParameter,
    StringParameter,
)

from ...model.ptychi import (
    PtyChiDMSettings,
    PtyChiEnumerators,
    PtyChiLSQMLSettings,
    PtyChiObjectSettings,
    PtyChiPIESettings,
)
from ..parametric import (
    CheckBoxParameterViewController,
    CheckableGroupBoxParameterViewController,
    ComboBoxParameterViewController,
    DecimalLineEditParameterViewController,
    DecimalSliderParameterViewController,
    LengthWidgetParameterViewController,
    SpinBoxParameterViewController,
)
from .optimizer import PtyChiOptimizationPlanViewController, PtyChiOptimizerParameterViewController

__all__ = ['PtyChiObjectViewController']


class PtyChiOptimizeSliceSpacingViewController(CheckableGroupBoxParameterViewController):
    def __init__(
        self,
        optimize_slice_spacing: BooleanParameter,
        start: IntegerParameter,
        stop: IntegerParameter,
        stride: IntegerParameter,
        optimizer: StringParameter,
        step_size: RealParameter,
        num_epochs: IntegerParameter,
        enumerators: PtyChiEnumerators,
    ) -> None:
        super().__init__(
            optimize_slice_spacing,
            'Optimize Slice Spacing',
            tool_tip='Whether to optimize the slice spacing',
        )
        self._plan_view_controller = PtyChiOptimizationPlanViewController(
            start,
            stop,
            stride,
            num_epochs,
        )
        self._optimizer_view_controller = PtyChiOptimizerParameterViewController(
            optimizer, enumerators
        )
        self._step_size_view_controller = DecimalLineEditParameterViewController(
            step_size, tool_tip='Optimizer step size'
        )

        layout = QFormLayout()
        layout.addRow('Plan:', self._plan_view_controller.get_widget())
        layout.addRow('Optimizer:', self._optimizer_view_controller.get_widget())
        layout.addRow('Step Size:', self._step_size_view_controller.get_widget())
        self.get_widget().setLayout(layout)


class PtyChiConstrainL1NormViewController(CheckableGroupBoxParameterViewController):
    def __init__(
        self,
        constrain_l1_norm: BooleanParameter,
        start: IntegerParameter,
        stop: IntegerParameter,
        stride: IntegerParameter,
        weight: RealParameter,
        num_epochs: IntegerParameter,
    ) -> None:
        super().__init__(
            constrain_l1_norm,
            'Constrain L\u2081 Norm',
            tool_tip='Whether to constrain the L\u2081 norm',
        )
        self._plan_view_controller = PtyChiOptimizationPlanViewController(
            start, stop, stride, num_epochs
        )
        self._weight_view_controller = DecimalLineEditParameterViewController(
            weight,
            tool_tip='Weight of the L\u2081 norm constraint',
        )

        layout = QFormLayout()
        layout.addRow('Plan:', self._plan_view_controller.get_widget())
        layout.addRow('Weight:', self._weight_view_controller.get_widget())
        self.get_widget().setLayout(layout)


class PtyChiConstrainL2NormViewController(CheckableGroupBoxParameterViewController):
    def __init__(
        self,
        constrain_l2_norm: BooleanParameter,
        start: IntegerParameter,
        stop: IntegerParameter,
        stride: IntegerParameter,
        weight: RealParameter,
        num_epochs: IntegerParameter,
    ) -> None:
        super().__init__(
            constrain_l2_norm,
            'Constrain L\u2082 Norm',
            tool_tip='Whether to constrain the L\u2082 norm',
        )
        self._plan_view_controller = PtyChiOptimizationPlanViewController(
            start, stop, stride, num_epochs
        )
        self._weight_view_controller = DecimalLineEditParameterViewController(
            weight,
            tool_tip='Weight of the L\u2082 norm constraint',
        )

        layout = QFormLayout()
        layout.addRow('Plan:', self._plan_view_controller.get_widget())
        layout.addRow('Weight:', self._weight_view_controller.get_widget())
        self.get_widget().setLayout(layout)


class PtyChiConstrainSmoothnessViewController(CheckableGroupBoxParameterViewController):
    def __init__(
        self,
        constrain_smoothness: BooleanParameter,
        start: IntegerParameter,
        stop: IntegerParameter,
        stride: IntegerParameter,
        alpha: RealParameter,
        num_epochs: IntegerParameter,
    ) -> None:
        super().__init__(
            constrain_smoothness,
            'Constrain Smoothness',
            tool_tip='Whether to constrain smoothness in the magnitude (but not phase) of the object',
        )
        self._plan_view_controller = PtyChiOptimizationPlanViewController(
            start, stop, stride, num_epochs
        )
        self._alpha_view_controller = DecimalSliderParameterViewController(
            alpha, tool_tip='Relaxation smoothing constant'
        )

        layout = QFormLayout()
        layout.addRow('Plan:', self._plan_view_controller.get_widget())
        layout.addRow('Alpha:', self._alpha_view_controller.get_widget())
        self.get_widget().setLayout(layout)


class PtyChiConstrainTotalVariationViewController(CheckableGroupBoxParameterViewController):
    def __init__(
        self,
        constrain_total_variation: BooleanParameter,
        start: IntegerParameter,
        stop: IntegerParameter,
        stride: IntegerParameter,
        weight: RealParameter,
        num_epochs: IntegerParameter,
    ) -> None:
        super().__init__(
            constrain_total_variation,
            'Constrain Total Variation',
            tool_tip='Whether to constrain the total variation',
        )
        self._plan_view_controller = PtyChiOptimizationPlanViewController(
            start, stop, stride, num_epochs
        )
        self._weight_view_controller = DecimalLineEditParameterViewController(
            weight,
            tool_tip='Weight of the total variation constraint',
        )

        layout = QFormLayout()
        layout.addRow('Plan:', self._plan_view_controller.get_widget())
        layout.addRow('Weight:', self._weight_view_controller.get_widget())
        self.get_widget().setLayout(layout)


class PtyChiRemoveGridArtifactsViewController(CheckableGroupBoxParameterViewController):
    def __init__(
        self,
        remove_grid_artifacts: BooleanParameter,
        start: IntegerParameter,
        stop: IntegerParameter,
        stride: IntegerParameter,
        period_x_m: RealParameter,
        period_y_m: RealParameter,
        window_size_px: IntegerParameter,
        direction: StringParameter,
        num_epochs: IntegerParameter,
        enumerators: PtyChiEnumerators,
    ) -> None:
        super().__init__(
            remove_grid_artifacts,
            'Remove Grid Artifacts',
            tool_tip="Whether to remove grid artifacts in the object's phase at the end of an epoch.",
        )
        self._plan_view_controller = PtyChiOptimizationPlanViewController(
            start, stop, stride, num_epochs
        )
        self._period_x_view_controller = LengthWidgetParameterViewController(
            period_x_m, tool_tip='Horizontal period of grid artifacts in meters'
        )
        self._period_y_view_controller = LengthWidgetParameterViewController(
            period_y_m, tool_tip='Vertical period of grid artifacts in meters'
        )
        self._window_size_view_controller = SpinBoxParameterViewController(
            window_size_px, tool_tip='Window size for grid artifact removal in pixels'
        )
        self._direction_view_controller = ComboBoxParameterViewController(
            direction, enumerators.directions(), tool_tip='Direction of grid artifact removal'
        )

        layout = QFormLayout()
        layout.addRow('Plan:', self._plan_view_controller.get_widget())
        layout.addRow('Period X:', self._period_x_view_controller.get_widget())
        layout.addRow('Period Y:', self._period_y_view_controller.get_widget())
        layout.addRow('Window Size [px]:', self._window_size_view_controller.get_widget())
        layout.addRow('Direction:', self._direction_view_controller.get_widget())
        self.get_widget().setLayout(layout)


class PtyChiRegularizeMultisliceViewController(CheckableGroupBoxParameterViewController):
    def __init__(
        self,
        regularize_multislice: BooleanParameter,
        start: IntegerParameter,
        stop: IntegerParameter,
        stride: IntegerParameter,
        weight: RealParameter,
        unwrap_phase: BooleanParameter,
        gradient_method: StringParameter,
        integration_method: StringParameter,
        num_epochs: IntegerParameter,
        enumerators: PtyChiEnumerators,
    ) -> None:
        super().__init__(
            regularize_multislice,
            'Regularize Multislice',
            tool_tip='Whether to regularize multislice objects using cross-slice smoothing',
        )
        self._plan_view_controller = PtyChiOptimizationPlanViewController(
            start, stop, stride, num_epochs
        )
        self._weight_view_controller = DecimalLineEditParameterViewController(
            weight,
            tool_tip='Weight for multislice regularization',
        )
        self._unwrap_phase_view_controller = CheckBoxParameterViewController(
            unwrap_phase,
            'Unwrap Phase',
            tool_tip='Whether to unwrap the phase of the object during multislice regularization',
        )
        self._gradient_method_view_controller = ComboBoxParameterViewController(
            gradient_method,
            enumerators.image_gradient_methods(),
            tool_tip='Method for calculating the phase gradient during phase unwrapping',
        )
        self._integration_method_view_controller = ComboBoxParameterViewController(
            integration_method,
            enumerators.image_integration_methods(),
            tool_tip='Method for integrating the phase gradient during phase unwrapping',
        )

        layout = QFormLayout()
        layout.addRow('Plan:', self._plan_view_controller.get_widget())
        layout.addRow('Weight:', self._weight_view_controller.get_widget())
        layout.addRow(self._unwrap_phase_view_controller.get_widget())
        layout.addRow('Gradient Method:', self._gradient_method_view_controller.get_widget())
        layout.addRow('Integration Method:', self._integration_method_view_controller.get_widget())
        self.get_widget().setLayout(layout)


class PtyChiRemoveObjectProbeAmbiguityViewController(CheckableGroupBoxParameterViewController):
    def __init__(
        self,
        remove_object_probe_ambiguity: BooleanParameter,
        start: IntegerParameter,
        stop: IntegerParameter,
        stride: IntegerParameter,
        num_epochs: IntegerParameter,
    ) -> None:
        super().__init__(
            remove_object_probe_ambiguity,
            'Remove Object Probe Ambiguity',
            tool_tip='Whether to remove object-probe ambiguity',
        )
        self._plan_view_controller = PtyChiOptimizationPlanViewController(
            start, stop, stride, num_epochs
        )

        layout = QFormLayout()
        layout.addRow('Plan:', self._plan_view_controller.get_widget())
        self.get_widget().setLayout(layout)


class PtyChiObjectViewController(CheckableGroupBoxParameterViewController):
    def __init__(
        self,
        settings: PtyChiObjectSettings,
        dm_settings: PtyChiDMSettings | None,
        lsqml_settings: PtyChiLSQMLSettings | None,
        pie_settings: PtyChiPIESettings | None,
        num_epochs: IntegerParameter,
        enumerators: PtyChiEnumerators,
    ) -> None:
        super().__init__(
            settings.is_optimizable,
            'Optimize Object',
            tool_tip='Whether the object is optimizable',
        )
        self._optimization_plan_view_controller = PtyChiOptimizationPlanViewController(
            settings.optimization_plan_start,
            settings.optimization_plan_stop,
            settings.optimization_plan_stride,
            num_epochs,
        )
        self._optimizer_view_controller = PtyChiOptimizerParameterViewController(
            settings.optimizer, enumerators
        )
        self._step_size_view_controller = DecimalLineEditParameterViewController(
            settings.step_size, tool_tip='Optimizer step size'
        )
        self._optimize_slice_spacing_view_controller = PtyChiOptimizeSliceSpacingViewController(
            settings.optimize_slice_spacing,
            settings.optimize_slice_spacing_start,
            settings.optimize_slice_spacing_stop,
            settings.optimize_slice_spacing_stride,
            settings.optimize_slice_spacing_optimizer,
            settings.optimize_slice_spacing_step_size,
            num_epochs,
            enumerators,
        )
        self._constrain_l1_norm_view_controller = PtyChiConstrainL1NormViewController(
            settings.constrain_l1_norm,
            settings.constrain_l1_norm_start,
            settings.constrain_l1_norm_stop,
            settings.constrain_l1_norm_stride,
            settings.constrain_l1_norm_weight,
            num_epochs,
        )
        self._constrain_l2_norm_view_controller = PtyChiConstrainL2NormViewController(
            settings.constrain_l2_norm,
            settings.constrain_l2_norm_start,
            settings.constrain_l2_norm_stop,
            settings.constrain_l2_norm_stride,
            settings.constrain_l2_norm_weight,
            num_epochs,
        )
        self._constrain_smoothness_view_controller = PtyChiConstrainSmoothnessViewController(
            settings.constrain_smoothness,
            settings.constrain_smoothness_start,
            settings.constrain_smoothness_stop,
            settings.constrain_smoothness_stride,
            settings.constrain_smoothness_alpha,
            num_epochs,
        )
        self._constrain_total_variation_view_controller = (
            PtyChiConstrainTotalVariationViewController(
                settings.constrain_total_variation,
                settings.constrain_total_variation_start,
                settings.constrain_total_variation_stop,
                settings.constrain_total_variation_stride,
                settings.constrain_total_variation_weight,
                num_epochs,
            )
        )
        self._remove_grid_artifacts_view_controller = PtyChiRemoveGridArtifactsViewController(
            settings.remove_grid_artifacts,
            settings.remove_grid_artifacts_start,
            settings.remove_grid_artifacts_stop,
            settings.remove_grid_artifacts_stride,
            settings.remove_grid_artifacts_period_x_m,
            settings.remove_grid_artifacts_period_y_m,
            settings.remove_grid_artifacts_window_size_px,
            settings.remove_grid_artifacts_direction,
            num_epochs,
            enumerators,
        )
        self._regularize_multislice_view_controller = PtyChiRegularizeMultisliceViewController(
            settings.regularize_multislice,
            settings.regularize_multislice_start,
            settings.regularize_multislice_stop,
            settings.regularize_multislice_stride,
            settings.regularize_multislice_weight,
            settings.regularize_multislice_unwrap_phase,
            settings.regularize_multislice_unwrap_phase_image_gradient_method,
            settings.regularize_multislice_unwrap_phase_image_integration_method,
            num_epochs,
            enumerators,
        )
        self._patch_interpolator_view_controller = ComboBoxParameterViewController(
            settings.patch_interpolator,
            enumerators.patch_interpolation_methods(),
            tool_tip='Interpolation method used for extracting and updating patches of the object',
        )
        self._remove_object_probe_ambiguity_view_controller = (
            PtyChiRemoveObjectProbeAmbiguityViewController(
                settings.remove_object_probe_ambiguity,
                settings.remove_object_probe_ambiguity_start,
                settings.remove_object_probe_ambiguity_stop,
                settings.remove_object_probe_ambiguity_stride,
                num_epochs,
            )
        )
        self._build_preconditioner_with_all_modes_view_controller = CheckBoxParameterViewController(
            settings.build_preconditioner_with_all_modes,
            'Build Preconditioner with All Modes',
            tool_tip='Whether to build the preconditioner using all modes',
        )

        layout = QFormLayout()
        layout.addRow('Plan:', self._optimization_plan_view_controller.get_widget())
        layout.addRow('Optimizer:', self._optimizer_view_controller.get_widget())
        layout.addRow('Step Size:', self._step_size_view_controller.get_widget())
        layout.addRow(self._optimize_slice_spacing_view_controller.get_widget())
        layout.addRow(self._constrain_l1_norm_view_controller.get_widget())
        layout.addRow(self._constrain_l2_norm_view_controller.get_widget())
        layout.addRow(self._constrain_smoothness_view_controller.get_widget())
        layout.addRow(self._constrain_total_variation_view_controller.get_widget())
        layout.addRow(self._remove_grid_artifacts_view_controller.get_widget())
        layout.addRow(self._regularize_multislice_view_controller.get_widget())
        layout.addRow('Patch Interpolator:', self._patch_interpolator_view_controller.get_widget())
        layout.addRow(self._remove_object_probe_ambiguity_view_controller.get_widget())
        layout.addRow(self._build_preconditioner_with_all_modes_view_controller.get_widget())

        if dm_settings is not None:
            self._amplitude_clamp_limit_view_controller = DecimalLineEditParameterViewController(
                dm_settings.object_amplitude_clamp_limit,
                tool_tip='Maximum amplitude value for the object',
            )
            layout.addRow(
                'Amplitude Clamp Limit:', self._amplitude_clamp_limit_view_controller.get_widget()
            )

            self._inertia_view_controller = DecimalLineEditParameterViewController(
                dm_settings.object_inertia,
                tool_tip='Inertia for the object update',
            )
            layout.addRow('Inertia:', self._inertia_view_controller.get_widget())

        if lsqml_settings is not None:
            self._object_optimal_step_size_scaler_view_controller = (
                DecimalLineEditParameterViewController(
                    lsqml_settings.object_optimal_step_size_scaler,
                    tool_tip='Optimal step size scaler for the object update',
                )
            )
            layout.addRow(
                'Optimal Step Size Scaler:',
                self._object_optimal_step_size_scaler_view_controller.get_widget(),
            )

            self._object_multimodal_update_view_controller = CheckBoxParameterViewController(
                lsqml_settings.object_multimodal_update,
                'Multimodal Update',
                tool_tip='When checked, the object update direction is calculated and summed over all probe modes rather than only the first mode',
            )
            layout.addRow(self._object_multimodal_update_view_controller.get_widget())

        if pie_settings is not None:
            self._alpha_view_controller = DecimalSliderParameterViewController(
                pie_settings.object_alpha,
                tool_tip='Relaxation factor for the object update',
            )
            layout.addRow('Alpha:', self._alpha_view_controller.get_widget())

        self.get_widget().setLayout(layout)



import numpy

from ptychodus.api.diffraction import BadPixels, DiffractionIndexes, DiffractionPatterns
from ptychodus.api.io import AssembledDiffractionData
from ptychodus.api.propagator import (
    FraunhoferPropagator,
    PropagatorParameters,
)

from ..diffraction import AssembledDiffractionDataset
from ..product import ProductRepository
from .interpolators import BarycentricArrayInterpolator


class DiffractionSimulator:
    def __init__(self, dataset: AssembledDiffractionDataset, repository: ProductRepository) -> None:
        super().__init__()
        self._dataset = dataset
        self._repository = repository

    def simulate(self, product_index: int) -> None:
        product = self._repository[product_index].get_product()
        # TODO also support multislice
        interpolator = BarycentricArrayInterpolator(product.object_.get_layer(0))
        probe_geometry = product.probes.get_geometry()

        propagator_parameters = PropagatorParameters(
            wavelength_m=product.metadata.probe_wavelength_m,
            width_px=probe_geometry.width_px,
            height_px=probe_geometry.height_px,
            pixel_width_m=probe_geometry.pixel_width_m,
            pixel_height_m=probe_geometry.pixel_height_m,
            propagation_distance_m=product.metadata.detector_distance_m,
        )

        # TODO also support near-field propagation
        propagator = FraunhoferPropagator(propagator_parameters)

        num_positions = len(product.probe_positions)
        indexes: DiffractionIndexes = numpy.array(num_positions, dtype=int)
        patterns: DiffractionPatterns = numpy.zeros(
            (num_positions, probe_geometry.height_px, probe_geometry.width_px),
            dtype=float,
        )
        bad_pixels: BadPixels = numpy.full(
            (probe_geometry.height_px, probe_geometry.width_px), False
        )

        for index, (probe_position, probe) in enumerate(
            zip(product.probe_positions, product.probes)
        ):
            object_geometry = product.object_.get_geometry()
            object_position = object_geometry.map_coordinates_probe_to_object(probe_position)

            probe_geometry = product.probes.get_geometry()
            object_patch = interpolator.get_patch(
                object_position.coordinate_x_px,
                object_position.coordinate_y_px,
                probe_geometry.width_px,
                probe_geometry.height_px,
            )

            for imode in range(probe.num_incoherent_modes):
                exit_wave = probe.get_incoherent_mode(imode) * object_patch
                wavefield = propagator.propagate(exit_wave)
                patterns[index, :, :] += numpy.square(numpy.abs(wavefield))

            indexes[index] = object_position.index

        data = AssembledDiffractionData(
            indexes=indexes,
            patterns=patterns,
            bad_pixels=bad_pixels,
        )
        self._dataset.set_assembled_patterns(data)



from PyQt5.QtWidgets import QFormLayout

from ptychodus.api.parametric import BooleanParameter, IntegerParameter, StringParameter

from ...model.ptychi import PtyChiEnumerators, PtyChiOPRSettings
from ..parametric import (
    CheckBoxParameterViewController,
    CheckableGroupBoxParameterViewController,
    ComboBoxParameterViewController,
    DecimalLineEditParameterViewController,
    DecimalSliderParameterViewController,
    SpinBoxParameterViewController,
)
from .optimizer import PtyChiOptimizationPlanViewController, PtyChiOptimizerParameterViewController


class PtyChiSmoothOPRModeWeightsViewController(CheckableGroupBoxParameterViewController):
    def __init__(
        self,
        smooth_mode_weights: BooleanParameter,
        start: IntegerParameter,
        stop: IntegerParameter,
        stride: IntegerParameter,
        smoothing_method: StringParameter,
        polynomial_smoothing_degree: IntegerParameter,
        num_epochs: IntegerParameter,
        enumerators: PtyChiEnumerators,
    ) -> None:
        super().__init__(
            smooth_mode_weights,
            'Smooth OPR Mode Weights',
            tool_tip='Smooth the OPR mode weights',
        )
        self._plan_view_controller = PtyChiOptimizationPlanViewController(
            start, stop, stride, num_epochs
        )
        self._smoothing_method_view_controller = ComboBoxParameterViewController(
            smoothing_method,
            enumerators.opr_weight_smoothing_methods(),
            tool_tip='Method for smoothing OPR mode weights',
        )
        self._polynomial_smoothing_degree_view_controller = SpinBoxParameterViewController(
            polynomial_smoothing_degree,
            tool_tip='Degree of the polynomial used for smoothing OPR mode weights',
        )

        layout = QFormLayout()
        layout.addRow('Plan:', self._plan_view_controller.get_widget())
        layout.addRow('Smoothing Method:', self._smoothing_method_view_controller.get_widget())
        layout.addRow(
            'Polynomial Degree:', self._polynomial_smoothing_degree_view_controller.get_widget()
        )
        self.get_widget().setLayout(layout)


class PtyChiOPRViewController(CheckableGroupBoxParameterViewController):
    def __init__(
        self,
        settings: PtyChiOPRSettings,
        num_epochs: IntegerParameter,
        enumerators: PtyChiEnumerators,
    ) -> None:
        super().__init__(
            settings.is_optimizable,
            'Orthogonal Probe Relaxation',
            tool_tip='Whether OPR modes are optimizable',
        )
        self._optimization_plan_view_controller = PtyChiOptimizationPlanViewController(
            settings.optimization_plan_start,
            settings.optimization_plan_stop,
            settings.optimization_plan_stride,
            num_epochs,
        )
        self._optimizer_view_controller = PtyChiOptimizerParameterViewController(
            settings.optimizer, enumerators
        )
        self._step_size_view_controller = DecimalLineEditParameterViewController(
            settings.step_size, tool_tip='Optimizer step size'
        )
        self._optimize_eigenmode_weights_view_controller = CheckBoxParameterViewController(
            settings.optimize_eigenmode_weights,
            'Optimize Eigenmode Weights',
            tool_tip='Whether to optimize eigenmode weights (i.e., the weights of the second and following OPR modes)',
        )
        self._optimize_intensities_view_controller = CheckBoxParameterViewController(
            settings.optimize_intensities,
            'Optimize Intensities',
            tool_tip='Whether to optimize intensity variation (i.e., the weight of the first OPR mode)',
        )
        self._smooth_mode_weights_view_controller = PtyChiSmoothOPRModeWeightsViewController(
            settings.smooth_mode_weights,
            settings.smooth_mode_weights_start,
            settings.smooth_mode_weights_stop,
            settings.smooth_mode_weights_stride,
            settings.smoothing_method,
            settings.polynomial_smoothing_degree,
            num_epochs,
            enumerators,
        )
        self._relax_update_view_controller = DecimalSliderParameterViewController(
            settings.relax_update,
            tool_tip='Whether to relax the update of the OPR mode weights',
        )

        layout = QFormLayout()
        layout.addRow('Plan:', self._optimization_plan_view_controller.get_widget())
        layout.addRow('Optimizer:', self._optimizer_view_controller.get_widget())
        layout.addRow('Step Size:', self._step_size_view_controller.get_widget())
        layout.addRow(self._optimize_intensities_view_controller.get_widget())
        layout.addRow(self._optimize_eigenmode_weights_view_controller.get_widget())
        layout.addRow(self._smooth_mode_weights_view_controller.get_widget())
        layout.addRow('Relax Update:', self._relax_update_view_controller.get_widget())
        self.get_widget().setLayout(layout)
